con' SPI pins
  SPI_SEL_CLK            = 8  ' used to clock the SPI counter that feeds 4-bit value to SPI selection decoder, clocks on rising edge
  SPI_SEL_CLR            = 25 ' used to clear the SPI counter back to 0000b (NULL SPI device), active LOW

  SPI_MOSI               = 9  ' SPI master out serial in to slave
  SPI_MISO               = 10 ' SPI master in serial out from slave, 10 for good C3, change to 7 for BROKEN C3 with YELLOW jumper.
  SPI_SCK                = 11 ' SPI clock from master to all slaves

  ' SPI channels defines 
  SPI_CHAN_NULL          = 0 ' NULL channel, disables all on/off board devices.
  SPI_CHAN_SRAM0         = 1 ' 32K SRAM Bank 0.
  SPI_CHAN_SRAM1         = 2 ' 32K SRAM Bank 1.
  SPI_CHAN_FLASH         = 3 ' 1MB FLASH Memory.
  SPI_CHAN_AD            = 4 ' MCP3202 2-Channel 12-bit A/D.
  SPI_CHAN_SD            = 5 ' Micro SD Card.

  ' interface header SPI selects
  SPI_CHAN_SPI6          = 6 ' Header Interface SPI6.
  SPI_CHAN_SPI7 



PUB SRAM_Init ( mode ) | _spi_word
{{
This function initializes both banks of the Microchip 23K256 SRAMs
serial SPI SRAMs'. The initialization simply sets the STATUS register
of each SRAM into "byte", "page" or "sequential" read/write mode to allow more
efficient access of the SRAMs. The SRAM API currently relies on "sequential"
mode, so if yoo want to use them, this function should be called with SPI_MODE_SEQUENTIAL

The 23K256 supports a number of commands, we have
only implemented a couple here in these API functions (single byte read/write,
multiple byte read/write). However, there are many others you can implement, find
out more by reviewing the chip specs and data sheets found here:

http://www.microchip.com/wwwproducts/Devices.aspx?dDocName=en539039
http://ww1.microchip.com/downloads/en/DeviceDoc/22100E.pdf

PARMS:  mode - %00_00000_1 = SPI_MODE_BYTE
               %10_00000_1 = SPI_MODE_PAGE
               %01_00000_1 = SPI_MODE_SEQUENTIAL

RETURNS: Nothing. 

}}

  ' first set SRAM bank 0 to sequential access mode (see page 6 of data sheet)
  SPI_Select_Channel( SPI_CHAN_SRAM0 )

  ' write status register command "sequential access mode" along with disable "HOLD" to setup SRAM for sequential reading/writing
  SPI_Write_Read( 16, %0000000_1 << 8 |  mode , $FF )

  ' now select SRAM bank 1
  SPI_Select_Channel( SPI_CHAN_SRAM1 )

  ' write status register command "sequential access mode" along with disable "HOLD" to setup SRAM for sequential reading/writing
  SPI_Write_Read( 16, %0000000_1 << 8 |  mode , $FF )

  ' and finally de-select all SPI devices
  SPI_Select_Channel( 0 )

' end SRAM_Init

' /////////////////////////////////////////////////////////////////////////////

PUB SRAM_Write( bank, addr, num_bytes, src_buffer_ptr) | _index, _data, _spi_word
{{
This function writes a number of bytes to either bank of the Microchip 23K256
serial SPI SRAMs' from a buffer. The 23K256 supports a number of commands, we have
only implemented a couple here in these API functions (single byte read/write,
multiple byte read/write). However, there are many others you can implement, find
out more by reviewing the chip specs and data sheets found here:

http://www.microchip.com/wwwproducts/Devices.aspx?dDocName=en539039
http://ww1.microchip.com/downloads/en/DeviceDoc/22100E.pdf


PARMS:  bank           - [0,1] the SRAM bank to write to, there are two banks, each holds 32K bytes
        addr           - address in SRAM to write to [0...32767]
        num_bytes      - number of bytes to write from source buffer to SRAM    
        src_buffer_ptr - pointer to the buffer of bytes to write to SRAM

RETURNS: number of bytes written 

}}

  ' test if there is anything to write?
  if (num_bytes => 1)

    ' select SPI channel for requested SRAM bank
    SPI_Select_Channel( SPI_CHAN_SRAM0 + bank )

    ' set instruction word ( write sequential byte:8 | address:16 )
    _spi_word := (%00000010 << 16) + (addr)

    ' and now send command to SRAM for sequential writing
    SPI_Write_Read( 24, _spi_word, $FF )
    
    ' write the bytes into the SRAM
    repeat _index from 0 to num_bytes-1

      ' get byte to write
      _data := byte[ src_buffer_ptr][ _index ]
       
      ' write the byte
      SPI_Write_Read( 8, _data, $FF )

    ' de-select SRAM SPI channel
    SPI_Select_Channel( 0 )

    ' return bytes written
    return( num_bytes )

  else
    ' catch error
    return 0 

' end SRAM_Write

' /////////////////////////////////////////////////////////////////////////////

PUB SRAM_Fill( bank, addr, num_bytes, value) | _index, _data, _spi_word
{{
This function writes a constant value to either bank of the Microchip 23K256
serial SPI SRAMs' from a buffer. The 23K256 supports a number of commands, we have
only implemented a couple here in these API functions (single byte read/write,
multiple byte read/write). However, there are many others you can implement, find
out more by reviewing the chip specs and data sheets found here:

http://www.microchip.com/wwwproducts/Devices.aspx?dDocName=en539039
http://ww1.microchip.com/downloads/en/DeviceDoc/22100E.pdf


PARMS:  bank           - [0,1] the SRAM bank to write to, there are two banks, each holds 32K bytes
        addr           - address in SRAM to write to [0...32767]
        num_bytes      - number of bytes to write from source buffer to SRAM    
        value          - byte value to write to SRAM

RETURNS: number of bytes written 

}}

  ' test if there is anything to fill?
  if (num_bytes => 1)

    ' select SPI channel for requested SRAM bank
    SPI_Select_Channel( SPI_CHAN_SRAM0 + bank )

    ' set instruction word ( write byte sequential:8 | address:16 )
    _spi_word := (%00000010 << 16) + (addr)

    ' setup for sequential byte write, this includes the command and address only
    SPI_Write_Read( 24, _spi_word, $FF )

    ' now the chip is ready to receive bytes (as many as we want to send)
    repeat _index from 0 to num_bytes-1
      SPI_Write_Read( 8, value, $FF )

    ' de-select SRAM channel
    SPI_Select_Channel( 0 )

    ' return bytes written
    return( num_bytes )

  else
    ' catch error
    return 0 

' end SRAM_Fill

' /////////////////////////////////////////////////////////////////////////////

PUB SRAM_Read(bank, addr, num_bytes, dest_buffer_ptr) | _index, _data, _spi_word
{{

This function reads a number of bytes from either bank of the SRAM into a buffer.

The 23K256 supports a number of commands, we have only implemented a couple here
in these API functions (single byte read/write, multiple byte read/write).
However, there are many others you can implement, find out more by reviewing
the chip specs and data sheets found here:

http://www.microchip.com/wwwproducts/Devices.aspx?dDocName=en539039
http://ww1.microchip.com/downloads/en/DeviceDoc/22100E.pdf

PARMS:  bank            - [0,1] the SRAM bank to read from, there are two banks, each holds 32K bytes
        addr            - address in SRAM to read from [0..32767]
        num_bytes       - number of bytes to read from SRAM into destination buffer
        dest_buffer_ptr - pointer to the buffer of bytes to read SRAM into

RETURNS: number of bytes read 

}}

  ' test for anything to read?
  if (num_bytes => 1)

    ' select SPI channel for requested SRAM bank
    SPI_Select_Channel( SPI_CHAN_SRAM0 + bank )

    ' set instruction word ( read byte sequential:8 | address:16 )
    _spi_word := (%00000011 << 16) + (addr)

    ' setup for sequential byte read, this includes the command and address only
    SPI_Write_Read( 24, _spi_word, $FF )

    ' now the chip is ready to be read from
    repeat _index from 0 to num_bytes-1

      ' read next byte from SRAM into receiving buffer
      byte [dest_buffer_ptr][_index] := SPI_Write_Read( 8, $00, $FF )

    ' de-select SRAM channel
    SPI_Select_Channel( 0 )
    
    ' return bytes read
    return( num_bytes )
  else
    ' catch error
    return 0 

' end SRAM_Read

' /////////////////////////////////////////////////////////////////////////////                                                                               

PUB SRAM64K_MREAD( addr ) | _spi_word, _data
{{
This function reads a single byte from either bank of the Microchip 23K256
serial SPI SRAMs'. This function is different from the SRAM_Read function
in that it accesses the two banks of SRAM as a contiguous region of 64K bytes
for you, and thus can be though of as a crude "array" access macro, rather
than a function. Ultimately, this should be converted to ASM along with
all the other important peripheral access methods.

The 23K256 supports a number of commands, we have
only implemented a couple here in these API functions (single byte read/write,
multiple byte read/write). However, there are many others you can implement, find
out more by reviewing the chip specs and data sheets found here:

http://www.microchip.com/wwwproducts/Devices.aspx?dDocName=en539039
http://ww1.microchip.com/downloads/en/DeviceDoc/22100E.pdf


PARMS:  addr - address in SRAMs to write to [0...65535].
    
RETURNS: The 8-bit data from the SRAM. 

}}

    ' the only tricky thing here is we need to take an address 0..65535 and then
    ' decide which bank it refers to and which byte in that bank, basically, some
    ' bit masking and shifts to accomplish this

    ' select SPI channel for requested SRAM bank
    SPI_Select_Channel( SPI_CHAN_SRAM0 + (addr >> 15) )

    ' set instruction word ( read byte sequential:8 | address:16 | dummy:8 )
    _spi_word := (%00000011 << 24) + (addr << 8) + 0 

    ' read the byte of data 
    _data := SPI_Write_Read( 32, _spi_word, $FF )

    ' de-select SRAM SPI channel
    SPI_Select_Channel( 0 )

    ' return data
    return ( _data )
    
' end SRAM64K_READ

' /////////////////////////////////////////////////////////////////////////////

PUB SRAM64K_MWRITE( addr, data ) | _spi_word
{{
This function writes a single byte to either bank of the Microchip 23K256
serial SPI SRAMs'. This function is different from the SRAM_Write|Fill functions
in that it accesses the two banks of SRAM as a contiguous region of 64K bytes
for you, and thus can be though of as a crude "array" access macro, rather
than a function. Ultimately, this should be converted to ASM along with
all the other important peripheral access methods.

The 23K256 supports a number of commands, we have
only implemented a couple here in these API functions (single byte read/write,
multiple byte read/write). However, there are many others you can implement, find
out more by reviewing the chip specs and data sheets found here:

http://www.microchip.com/wwwproducts/Devices.aspx?dDocName=en539039
http://ww1.microchip.com/downloads/en/DeviceDoc/22100E.pdf


PARMS:  addr - address in SRAMs to write to [0...65535].
        data - the byte to write.    

RETURNS: Nothing. 

}}

    ' the only tricky thing here is we need to take an address 0..65535 and then
    ' decide which bank it refers to and which byte in that bank, basically, some
    ' bit masking and shifts to accomplish this

    ' select SPI channel for requested SRAM bank
    SPI_Select_Channel( SPI_CHAN_SRAM0 + (addr >> 15) )

    ' set instruction word ( write sequential byte:8 | address:16 | data: 8)
    _spi_word := (%00000010 << 24) + ( (addr & $7FFF) << 8) + data

    ' write the byte of data 
    SPI_Write_Read( 32, _spi_word, $FF )

    ' de-select SRAM SPI channel
    SPI_Select_Channel( 0 )

' end SRAM64K_WRITE

PUB SPI_Init
{{
This function initializes the SPI IO's, counter, and mux, selects channel 0 and returns.

PARMS: none.
RETURNS: nothing.
}}
  ' set SPI mux counter IOs up
  DIRA[SPI_SEL_CLK] := 1 ' set to output
  OUTA[SPI_SEL_CLK] := 0

  DIRA[SPI_SEL_CLR] := 1 ' set to output
  OUTA[SPI_SEL_CLR] := 0 ' CLR counter
  OUTA[SPI_SEL_CLR] := 1 ' allow counting

  ' set up SPI lines
  OUTA[SPI_MOSI]     := 0 ' set to LOW
  OUTA[SPI_SCK]      := 0 ' set to LOW
 
  DIRA[SPI_MOSI]     := 1 ' set to output
  DIRA[SPI_MISO]     := 0 ' set to input
  DIRA[SPI_SCK]      := 1 ' set to output

  ' set SPI to NULL channel 0
  SPI_Select_Channel( 0 )

' end SPI_Init

' /////////////////////////////////////////////////////////////////////////////

PUB SPI_Write_Read( num_bits, data_out, bit_mask) | data_in
{{
This function writes and reads SPI data a bit at a time (SPI is a circular buffer protocal), the data is in MSB to LSB format
and up to 32-bits can be transmitted and received, the final result is bit masked by bit_mask

PARMS:

num_bits : number of bits to transmit from data_out
data_out : source of data to transmit
bit_mask : final result of SPI transmission is masked with this to grab the relevant least significant bits

RETURNS: data retrieved from SPI transmission

}}
  ' clear result
  data_in := 0

  ' now read the bits in/out
  repeat num_bits
    ' drop clock
    OUTA[SPI_SCK] := 0

    ' place next bit on MOSI
    OUTA[SPI_MOSI] := ((data_out >> (num_bits-- - 1)) & $01)
 
    ' now read next bit from MISO
    data_in := (data_in << 1) + INA[SPI_MISO]
      
    ' raise clock
    OUTA[SPI_SCK] := 1

  ' set clock and MOSI to LOW on exit
  OUTA[SPI_MOSI]  := 0
  OUTA[SPI_SCK]   := 0

  ' at this point, the data has been written and read, return result
  return ( data_in & bit_mask )            
    
' end SPI_Write_Read

' /////////////////////////////////////////////////////////////////////////////

PUB SPI_Select_Channel( channel )
{{
This function sets the active SPI channel chip select on the SPI mux, this is accomplished by first resetting the
SPI counter that feeds the SPI select decoder, then upcounting the requested number of channels.

PARMS:

channel : channel 0 to 7 to enable where the channels are defined as follows
          0 - NULL channel, disables all on/off board devices.
          1 - 32K SRAM Bank 0.
          2 - 32K SRAM Bank 1.
          3 - 1MB FLASH Memory.
          4 - MCP3202 2-Channel 12-bit A/D.
          5 - Micro SD Card.
          6 - Header Interface SPI6.
          7 - Header Interface SPI7.

RETURNS: nothing.

}}

  ' requesting channel 0? If so, easy reset
  if (channel == 0)
    ' clear the 161
    OUTA[SPI_SEL_CLR] := 0 ' CLR counter
    OUTA[SPI_SEL_CLR] := 1 ' allow counting
    return

  ' else non-null channel, count up to channel...
  
  ' first reset the SPI channel counter
  ' clear the 161
  OUTA[SPI_SEL_CLR] := 0 ' CLR counter
  OUTA[SPI_SEL_CLR] := 1 ' allow counting

  ' now increment to requested channel
  ' clock the 161
   OUTA[SPI_SEL_CLK] := 0

   repeat channel
     OUTA[SPI_SEL_CLK] := 1
     OUTA[SPI_SEL_CLK] := 0        

' end SPI_Select_Channel

DAT

  sram_buffer byte 1,4,9,16,25,36,49,64,81,100,121,144,169,196,225  